#!/usr/bin/env bash
## Test for issue1640 - <SYNOPSIS: darcs Issue 1640 darcs apply
## --verbose with no file args should mention stdin>
##
## Copyright (C) 2011  Radoslav Dorcik <dixiecko@gmail.com>
##
## Permission is hereby granted, free of charge, to any person
## obtaining a copy of this software and associated documentation
## files (the "Software"), to deal in the Software without
## restriction, including without limitation the rights to use, copy,
## modify, merge, publish, distribute, sublicense, and/or sell copies
## of the Software, and to permit persons to whom the Software is
## furnished to do so, subject to the following conditions:
##
## The above copyright notice and this permission notice shall be
## included in all copies or substantial portions of the Software.
##
## THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
## EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
## MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
## NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
## BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
## ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
## CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
## SOFTWARE.

. lib                           # Load some portability helpers.

#######################################################
# Test preparation 
#######################################################

# Create repository
darcs init      --repo S
darcs init      --repo T
darcs init      --repo T2
cd S
touch foo bar
darcs add foo bar
darcs record -a -m add_foo_bar -A x
darcs send --author=me --output=funpatch -a ../T


#######################################################
# Apply from stdin and check message
#######################################################

# Message when --verbose
cd ../T
darcs apply --verbose < ../S/funpatch | tee output.txt
grep "reading patch bundle from stdin..." output.txt || exit 1


# No message when no --verbose
cd ../T2
darcs apply < ../S/funpatch | tee output.txt
grep "reading patch bundle from stdin..." output.txt && exit 1
exit 0;
